"""
Python API for sequence distance functions.

"""

import abc
import warnings
from typing import (
    Any,
    Dict,
    Generic,
    List,
    Optional,
    Sequence,
    Set,
    TypeVar,
    Union,
)

import numpy as np
import numpy.typing as npt
import pandas as pd
from glom import glom
from scipy import spatial
from sklearn import preprocessing

import setriq._C as C

from .substitution import BLOSUM45, SubstitutionMatrix
from .utils import (
    TCR_DIST_DEFAULT,
    TcrDistDef,
    check_jaro_weights,
    check_jaro_winkler_params,
    enforce_list,
    ensure_equal_sequence_length,
)

__all__ = [
    "CdrDist",
    "Levenshtein",
    "TcrDist",
    "TcrDistComponent",
    "Hamming",
    "Jaro",
    "JaroWinkler",
    "LongestCommonSubstring",
    "OptimalStringAlignment",
]

FloatArray = npt.NDArray[np.float64]
SeqRecord = TypeVar("SeqRecord", bound=Union[str, Dict[str, str]])


class Metric(abc.ABC, Generic[SeqRecord]):
    """
    The Metric abstract base class. Users familiar with the torch paradigm will recognize the overall structure of
    the metric subclasses.

    Methods
    -------
    forward(self, *args, **kwargs):
        an abstract method which needs to be implemented in every subclass. It is accessed via the ``__call__`` method
        of the base class.

    """

    call_args: Dict[str, Any]

    def __init__(self, return_squareform: bool = False):
        self.return_squareform = return_squareform

    @abc.abstractmethod
    def forward(self, sequences: Sequence[SeqRecord]) -> List[float]:
        pass

    @enforce_list(argnum=1, convert_iterable=True)
    def __call__(self, sequences: Sequence[SeqRecord]) -> FloatArray:
        out = np.array(self.forward(sequences))
        if self.return_squareform:
            out = spatial.distance.squareform(out)

        return out

    def to_sklearn(self) -> preprocessing.FunctionTransformer:
        """Creates a FunctionTransformer from a given Metric instance.

        This is useful for integrating the distance functions into scikit-learn pipelines.

        Returns
        -------
        The FunctionTransformer which can be used with other scikit-learn tooling.

        """
        return preprocessing.FunctionTransformer(
            func=self,
            check_inverse=False,
        )


class CdrDist(Metric[str]):
    """
    The CdrDist [1]_ class. Inherits from Metric.

    Examples
    --------
    >>> sequences = ['CASSLKPNTEAFF', 'CASSAHIANYGYTF', 'CASRGATETQYF']
    >>>
    >>> metric = CdrDist()
    >>> distances = metric(sequences)

    References
    ----------
    .. [1] Thakkar, N. and Bailey-Kellogg, C., 2019. Balancing sensitivity and specificity in distinguishing TCR groups
       by CDR sequence similarity. BMC bioinformatics, 20(1), pp.1-14. (https://doi.org/10.1186/s12859-019-2864-8)

    """

    def __init__(
        self,
        substitution_matrix: SubstitutionMatrix = BLOSUM45,
        gap_opening_penalty: float = 10.0,
        gap_extension_penalty: float = 1.0,
        return_squareform: bool = False,
    ):
        super(CdrDist, self).__init__(return_squareform)
        self.call_args = {
            **substitution_matrix,
            "gap_opening_penalty": gap_opening_penalty,
            "gap_extension_penalty": gap_extension_penalty,
        }
        self.fn = C.cdr_dist

    def forward(self, sequences: Sequence[str]) -> List[float]:
        out = self.fn(sequences, **self.call_args)

        return out


class Levenshtein(Metric[str]):
    """
    The Levenshtein [1]_ class. Inherits from Metric. It uses a refactor of the ``python-Levenshtein`` [2]_
    implementation in the backend.

    Examples
    --------
    >>> sequences = ['CASSLKPNTEAFF', 'CASSAHIANYGYTF', 'CASRGATETQYF']
    >>>
    >>> metric = Levenshtein()
    >>> distances = metric(sequences)

    References
    ----------
    .. [1] Levenshtein, V.I., 1966, February. Binary codes capable of correcting deletions, insertions, and reversals.
       In Soviet physics doklady (Vol. 10, No. 8, pp. 707-710).
    .. [2] python-Levenshtein (https://github.com/ztane/python-Levenshtein)

    """

    def __init__(self, extra_cost: float = 0.0, return_squareform: bool = False):
        super(Levenshtein, self).__init__(return_squareform)
        self.call_args = {"extra_cost": extra_cost}
        self.fn = C.levenshtein

    def forward(self, sequences: Sequence[str]) -> List[float]:
        out = self.fn(sequences, **self.call_args)

        return out


class TcrDistComponent(Metric[str]):
    """
    The TcrDistComponent class. Inherits from Metric.

    Examples
    --------
    >>> from setriq import BLOSUM62
    >>> sequences = ['CASSLKPNTE', 'CASS-HIANY', 'CASRGAT--Q']
    >>>
    >>> metric = TcrDistComponent(substitution_matrix=BLOSUM62, gap_penalty=4., gap_symbol='-', weight=1.)
    >>> distances = metric(sequences)

    """

    def __init__(
        self,
        substitution_matrix: SubstitutionMatrix,
        gap_penalty: float,
        gap_symbol: str = "-",
        weight: float = 1.0,
        return_squareform: bool = False,
    ):
        """
        Initialize a TcrDistComponent object.

        Parameters
        ----------
        substitution_matrix : SubstitutionMatrix
            a SubstitutionMatrix object
        gap_penalty : float
            the gap penalty, i.e. the score given for a <not-gap-symbol> -> <gap-symbol> substitution
        gap_symbol : str
            the gap symbol (default = '-')
        weight : float
            the weighting of the component weight

        """
        super(TcrDistComponent, self).__init__(return_squareform)
        self.call_args = {
            **substitution_matrix,
            "gap_penalty": gap_penalty,
            "gap_symbol": gap_symbol,
            "weight": weight,
        }
        self.fn = C.tcr_dist_component

    @ensure_equal_sequence_length(argnum=1)
    def forward(self, sequences: Sequence[str]) -> List[float]:
        out = self.fn(sequences, **self.call_args)

        return out


class TcrDist(Metric[Dict[str, str]]):
    """
    TcrDist [1]_ class. Inherits from Metric. It is a container class for individual TcrDistComponent instances.
    Components are executed sequentially and their results aggregated at the end (summation).

    Attributes
    ----------
    components : List[str]
        holds the names of the components to be executed

    Examples
    --------
    >>> sequences = [
    ...     {'cdr_1': 'TSG------FNG', 'cdr_2': 'VVL----DGL', 'cdr_2_5': 'SRSN-GY', 'cdr_3': 'CAVR-----'},
    ...     {'cdr_1': 'TSG------FYG', 'cdr_2': 'NGL----DGL', 'cdr_2_5': 'SRSD-SY', 'cdr_3': 'CA-------'},
    ...     {'cdr_1': 'NSA------FQY', 'cdr_2': 'TYS----SGN', 'cdr_2_5': 'DKSSKY-', 'cdr_3': 'CAMS-----'}
    ... ]
    >>> metric = TcrDist()  # will produce a warning stating default configuration (Dash et al)
    >>> distances = metric(sequences)

    References
    ----------
    .. [1] Dash, P., Fiore-Gartland, A.J., Hertz, T., Wang, G.C., Sharma, S., Souquette, A., Crawford, J.C., Clemens,
       E.B., Nguyen, T.H., Kedzierska, K. and La Gruta, N.L., 2017. Quantifiable predictive features define
       epitope-specific T cell receptor repertoires. Nature, 547(7661), pp.89-93. (https://doi.org/10.1038/nature22383)

    """

    _default: TcrDistDef = TCR_DIST_DEFAULT
    _default_msg: str = (
        "TcrDist has been initialized using the default configuration. "
        "Please ensure that the input is a list of dictionaries, each with keys: {}"
    ).format(", ".join(repr(key) for key, _ in _default))

    def __init__(self, return_squareform: bool = False, **components: TcrDistComponent):
        """
        Initialize a TcrDist object. Initialization can happen in two ways:
            1. no arguments are passed, instantiating the default configuration of TcrDist -- i.e. the configuration
               described in Dash et al.
            2. a set of keyword arguments is passed, where each value is a TcrDistComponent instance and the key gives
               the component its name. An arbitrary number of components can be set and their naming is also arbitrary.

        Parameters
        ----------
        components : keyword arguments
            either a set of keyword arguments, where each value is a TcrDistComponent instance which will be stored as
            an attribute with the key as its name OR `None` -- in which case the default configuration is loaded (Dash
            et al).

        Examples
        --------

        Initialize with default parameters

        >>> metric = TcrDist()  # will produce a warning

        Initialize with custom components

        >>> component_1 = TcrDistComponent(substitution_matrix=BLOSUM45, gap_penalty=3, weight=1)
        >>> component_2 = TcrDistComponent(substitution_matrix=BLOSUM45, gap_penalty=5, weight=10)
        >>>
        >>> metric = TcrDist(cmp_1=component_1, cmp_2=component_2)

        Keep in mind that the keys will be used to assiociate the components to the relevant input fields, i.e. in this
        case the input should take the shape of:

        >>> [{'cmp_1': '...', 'cmp_2': '...'}, ...]

        additional keys will have no effect.

        """
        super(TcrDist, self).__init__(return_squareform)
        parts: List[str] = []

        # user-defined configuration
        if components:
            for name, component in components.items():
                # some type checking
                if not isinstance(component, TcrDistComponent):
                    raise TypeError(
                        f"{name!r} is not of type {TcrDistComponent.__class__.__name__}"
                    )

                setattr(self, name, component)
                parts.append(name)

        # default configuration
        else:
            for name, definition in self._default:
                setattr(self, name, TcrDistComponent(**definition))  # type: ignore[arg-type]
                parts.append(name)

            # warn user that default has been initialised and inform required input format
            warnings.warn(self._default_msg, UserWarning)

        self.components = parts

    def _check_input_format(self, ipt: pd.Index) -> None:
        pts: Set[str] = set(self.components)

        diff: Set[str] = pts.difference(ipt)
        if diff:
            raise ValueError("Missing key(s): {}".format(", ".join(map(repr, diff))))

    @property
    def required_input_keys(self) -> List[str]:
        """
        Get the keys (=fields) required in the input to TcrDist instance.

        Returns
        -------
        required_input_keys : List[str]
            returns a list of strings signifying the keys required in the input

        """
        return self.components

    @property
    def default_definition(self) -> TcrDistDef:
        """
        Get the default TcrDistComponent schema as defined by Dash et al.

        Returns
        -------
        default_schema : List[tuple]
            returns the schema for the TcrDistComponent instances held in the default configuration

        """
        return self._default

    def forward(self, sequences: Sequence[Dict[str, str]]) -> List[float]:
        if not sequences:
            return []

        # check the input keys provided -- assumes consistency
        self._check_input_format(pd.DataFrame(sequences).columns)

        # iterate through components and collect component output
        out: List[FloatArray] = []
        for part in self.components:
            # gather sequences of the associated field into a list
            sqs: List[str] = glom(sequences, [part])
            component: TcrDistComponent = getattr(self, part)

            # execute component on list of associated sequences
            result: FloatArray = component(sqs)
            out.append(result)

        # aggregate the component outputs
        res: FloatArray = np.stack(out).sum(axis=0)
        return res.tolist()


class Hamming(Metric[str]):
    """
    Hamming distance class. Inherits from Metric. Sequences must be of equal length.

    Examples
    --------
    >>> metric = Hamming(mismatch_score=2.0)
    >>> sequences = ['CASSLKPNTEAFF', 'CASSAHIANYGYTF', 'CASRGATETQYF']
    >>> distances = metric(sequences)

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Hamming_distance
    """

    def __init__(self, mismatch_score: float = 1.0, return_squareform: bool = False):
        super(Hamming, self).__init__(return_squareform)
        self.call_args = {"mismatch_score": mismatch_score}
        self.fn = C.hamming

    @ensure_equal_sequence_length(argnum=1)
    def forward(self, sequences: Sequence[str]) -> List[float]:
        out = self.fn(sequences, **self.call_args)
        return out


class Jaro(Metric[str]):
    """
    Jaro distance class. Inherits from Metric. Adapted from [2].

    Examples
    --------
    >>> metric = Jaro()
    >>> sequences = ['CASSLKPNTEAFF', 'CASSAHIANYGYTF', 'CASRGATETQYF']
    >>> distances = metric(sequences)

    References
    ----------
    [1] Jaro, M.A., 1989. Advances in record-linkage methodology as applied to matching the 1985 census of Tampa,
        Florida. Journal of the American Statistical Association, 84(406), pp.414-420.
    [2] Van der Loo, M.P., 2014. The stringdist package for approximate string matching. R J., 6(1), p.111.
    """

    def __init__(
        self,
        jaro_weights: Optional[List[float]] = None,
        return_squareform: bool = False,
    ):
        super(Jaro, self).__init__(return_squareform)
        jaro_weights = check_jaro_weights(jaro_weights)
        self.call_args = {"jaro_weights": jaro_weights}
        self.fn = C.jaro

    def forward(self, sequences: Sequence[str]) -> List[float]:
        out = self.fn(sequences, **self.call_args)
        return out


class JaroWinkler(Jaro):
    """
    Jaro-Winkler distance class. Inherits from Jaro.

    Examples
    --------
    >>> metric = JaroWinkler(p=0.10)
    >>> sequences = ['CASSLKPNTEAFF', 'CASSAHIANYGYTF', 'CASRGATETQYF']
    >>> distances = metric(sequences)

    References
    ----------
    .. [1] Winkler, W.E., 1990. String comparator metrics and enhanced decision rules in the Fellegi-Sunter model of
       record linkage.

    """

    @check_jaro_winkler_params
    def __init__(
        self,
        p: float = 0.10,
        max_l: int = 4,
        jaro_weights: Optional[List[float]] = None,
        return_squareform: bool = False,
    ):
        super(JaroWinkler, self).__init__(
            jaro_weights=jaro_weights, return_squareform=return_squareform
        )
        self.call_args["p"] = p
        self.call_args["max_l"] = max_l
        self.fn = C.jaro_winkler  # type: ignore[assignment]


class LongestCommonSubstring(Metric[str]):
    """
    Longest common substring distance class. Inherits from Metric.

    Examples
    --------
    >>> metric = LongestCommonSubstring()
    >>> sequences = ['CASSLKPNTEAFF', 'CASSAHIANYGYTF', 'CASRGATETQYF']
    >>> distances = metric(sequences)

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Longest_common_substring_problem

    """

    def __init__(self, return_squareform: bool = False):
        super(LongestCommonSubstring, self).__init__(return_squareform)
        self.fn = C.longest_common_substring

    def forward(self, sequences: Sequence[str]) -> List[float]:
        out = self.fn(sequences)
        return out


class OptimalStringAlignment(Metric[str]):
    """
    Optimal string alignment distance class. Inherits from Metric.

    Examples
    --------
    >>> metric = OptimalStringAlignment()
    >>> sequences = ['CASSLKPNTEAFF', 'CASSAHIANYGYTF', 'CASRGATETQYF']
    >>> distances = metric(sequences)

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Damerau%E2%80%93Levenshtein_distance

    """

    def __init__(self, return_squareform: bool = False):
        super(OptimalStringAlignment, self).__init__(return_squareform)
        self.fn = C.optimal_string_alignment

    def forward(self, sequences: Sequence[str]) -> List[float]:
        out = self.fn(sequences)
        return out



"""
modules
=======

This sub-package contains all of the `setriq` Python API.

Examples
--------
>>> from setriq import modules
>>> sub_mat = modules.BLOSUM62
>>> sub_mat('A', 'L')
... -1

"""

from . import single_dispatch
from .distances import (
    CdrDist,
    Hamming,
    Jaro,
    JaroWinkler,
    Levenshtein,
    LongestCommonSubstring,
    OptimalStringAlignment,
    TcrDist,
)
from .substitution import BLOSUM45, BLOSUM62, BLOSUM90, SubstitutionMatrix

__all__ = [
    "SubstitutionMatrix",
    "BLOSUM45",
    "BLOSUM62",
    "BLOSUM90",
    "CdrDist",
    "Levenshtein",
    "TcrDist",
    "Hamming",
    "Jaro",
    "JaroWinkler",
    "LongestCommonSubstring",
    "OptimalStringAlignment",
    "single_dispatch",
]



"""
Substitution matrix convenience interface.

"""

import abc
import copy
import pathlib
from typing import Any, Dict, List, Tuple, Union

import pkg_resources as pkg
import srsly

__all__ = [
    "BLOSUM45",
    "BLOSUM62",
    "BLOSUM90",
    "SubstitutionMatrix",
]


class SubstitutionMatrix(abc.ABC):
    r"""
    The SubstitutionMatrix abstract base class. It holds convenience methods for loading and using the classic
    biological sequence substitution matrices (e.g. ``BLOSUM``).

    Attributes
    ----------
    index : Dict[str, int]
        a mapping of strings (amino acids) to integers (matrix index)
    substitution_matrix : List[List[float]]
        the substitution scoring (:math:`N \times N`) matrix

    Methods
    -------
    from_json(self, file_path: Union[str, Path]) -> SubstitutionMatrix:
        load a substitution matrix from a json file

    Examples
    --------

    suppose we have a token index `idx` and a substitution matrix `scores`

    >>> idx = {'hello': 0, 'world': 1}
    >>> scores = [[1., -1.],
    ...           [-1., 1.]]
    >>> matrix = SubstitutionMatrix(index=idx, substitution_matrix=scores)

    here we can see that we can provide any arbitrary substitution matrix, but in general it is advised to use the
    pre-loaded BLOSUM matrices

    >>> [BLOSUM45, BLOSUM62, BLOSUM90]  # choose one of the following

    these are just instances of ``SubstitutionMatrix``, initialised through ``from_json``

    """
    _required_keys: Tuple[str, str] = (
        "index",
        "substitution_matrix",
    )
    index: Dict[str, int]
    substitution_matrix: List[List[float]]

    def __init__(
        self,
        index: Dict[str, int],
        substitution_matrix: List[List[float]],
        *args,
        **kwargs,
    ):
        """
        Construct an instance of a SubstitutionMatrix object.

        Parameters
        ----------
        index : Dict[str, int]
            a token index which maps the string tokens to their index in the matrix
        substitution_matrix : List[List[float]]
            the substitution scoring matrix
        args
        kwargs
        """
        self.index = index
        self.substitution_matrix = substitution_matrix

    @classmethod
    def from_json(cls, file_path: Union[str, pathlib.Path]) -> "SubstitutionMatrix":
        """
        Load a SubstitutionMatrix from a json file.

        Parameters
        ----------
        file_path : Union[str, Path]
            a path to a json file holding at least the token index and the substitution scoring matrix

        Returns
        -------
        substitution_matrix : SubstitutionMatrix
            returns an instance of the SubstitutionMatrix class, holding the values found at ``file_path``

        Examples
        --------
        >>> sub_mat = SubstitutionMatrix.from_json('/path/to/file.json')

        """
        if isinstance(file_path, str):
            file_path = pathlib.Path(file_path)

        values: Dict[str, Any] = srsly.read_json(file_path)
        for key in cls._required_keys:
            if key not in values:
                raise ValueError(
                    f"required key {repr(key)} not in provided file: {file_path}"
                )

        model = cls(**values)
        return model

    def __len__(self) -> int:
        return len(self.substitution_matrix)

    def __getitem__(self, key: str) -> Any:
        return getattr(self, key)

    def keys(self) -> Tuple[str, str]:
        return self._required_keys

    def __call__(self, a: str, b: str) -> float:
        """
        Get the score for a given substitution.

        Parameters
        ----------
        a : str
            a token in the index which substitutes
        b : str
            another token in the index which is substituted

        Returns
        -------
        substitution_score : float
            returns the score of a given substitution

        Examples
        --------
        >>> sub_mat = BLOSUM62
        >>> sub_mat('A', 'R')
        ... -1

        """
        i = self.index[a]
        j = self.index[b]

        out = self.substitution_matrix[i][j]
        return out

    def add_token(
        self, token: str, values: Union[float, List[float]], inplace: bool = False
    ) -> Union["SubstitutionMatrix", None]:
        """
        Add a special token to the substitution matrix with a given value or list of values.

        Parameters
        ----------
        token : str
            a special token to be added.
        values : Union[float, List[float]]
            a value or list of values to which the token will correspond. If a list of floats is provided, the list must
            have a length of ``len(substitution_matrix) + 1``, i.e. there must be number of rows + 1 elements in the
            list.
        inplace : bool
            boolean whether to add token inplace.

        Returns
        -------
        None
            this is an inplace operation

        Examples
        --------
        Single value example. The value is repeated to fit the required shape

        >>> sm = BLOSUM62
        >>> sm.add_token('-', 4.)

        List of floats example

        >>> sm = BLOSUM62
        >>> len(sm)
        ... 24
        >>> sm.add_token('setriq', [*range(26)])  # ints implicitly converted to floats

        """
        if token in self.index:
            raise ValueError("`token` already exists")

        # generate the full row in case of single value
        if isinstance(values, float):
            values = [values for _ in range(len(self.substitution_matrix[0]) + 1)]

        # check if first dimension fits
        if len(values) - 1 != len(self.substitution_matrix):
            raise ValueError(
                "`values` and `substitution_matrix` must have same dimension 0"
            )

        # generate deep copies to make inplace operation optional
        substitution_matrix = copy.deepcopy(self.substitution_matrix)
        index = copy.deepcopy(self.index)

        # append the values in the new row to the each existing row to enforce symmetry
        values = [*map(float, values)]
        for scoring_row, new_value in zip(substitution_matrix, values[:-1]):
            scoring_row.append(new_value)

        # append the new row and add new token
        substitution_matrix.append(values)
        index[token] = len(self.substitution_matrix) - 1

        if inplace:
            self.substitution_matrix = substitution_matrix
            self.index = index
            return None

        out = SubstitutionMatrix(index=index, substitution_matrix=substitution_matrix)
        return out


# below we load the matrices which come with the package -- this exposes them to the user
# they will be used for default settings in a number of metrics
PKG_NAME = __name__.split(".")[0]
DATA_DIR = pathlib.Path(pkg.resource_filename(PKG_NAME, "data/"))

BLOSUM45 = SubstitutionMatrix.from_json(DATA_DIR / "blosum-45.json")
BLOSUM62 = SubstitutionMatrix.from_json(DATA_DIR / "blosum-62.json")
BLOSUM90 = SubstitutionMatrix.from_json(DATA_DIR / "blosum-90.json")



"""
setriq
======

Provides:
    1. fast computation of pairwise (immunoglobulin) sequence distances
    2. parsed BLOSUM substitution matrices

About
-----
This package is a one-trick-pony which does its trick pretty well: efficient computation of pairwise sequence distances
on CPU. There are a number of distance functions available [1]_ [2]_ [3]_ [4]_, implemented in C++ and surfaced in
Python.

Examples
--------
>>> import setriq
>>> metric = setriq.CdrDist()
>>>
>>> sequences = [
...     'CASSLKPNTEAFF',
...     'CASSAHIANYGYTF',
...     'CASRGATETQYF'
... ]
>>> distances = metric(sequences)

References
----------
.. [1] Dash, P., Fiore-Gartland, A.J., Hertz, T., Wang, G.C., Sharma, S., Souquette, A., Crawford, J.C., Clemens, E.B.,
   Nguyen, T.H., Kedzierska, K. and La Gruta, N.L., 2017. Quantifiable predictive features define epitope-specific
   T cell receptor repertoires. Nature, 547(7661), pp.89-93. (https://doi.org/10.1038/nature22383)

.. [2] Levenshtein, V.I., 1966, February. Binary codes capable of correcting deletions, insertions, and reversals. In
   Soviet physics doklady (Vol. 10, No. 8, pp. 707-710).

.. [3] python-Levenshtein (https://github.com/ztane/python-Levenshtein)

.. [4] Thakkar, N. and Bailey-Kellogg, C., 2019. Balancing sensitivity and specificity in distinguishing TCR groups by
   CDR sequence similarity. BMC bioinformatics, 20(1), pp.1-14. (https://doi.org/10.1186/s12859-019-2864-8)

"""

from .modules import (
    BLOSUM45,
    BLOSUM62,
    BLOSUM90,
    CdrDist,
    Hamming,
    Jaro,
    JaroWinkler,
    Levenshtein,
    LongestCommonSubstring,
    OptimalStringAlignment,
    SubstitutionMatrix,
    TcrDist,
    single_dispatch,
)

__all__ = [
    "BLOSUM45",
    "BLOSUM62",
    "BLOSUM90",
    "CdrDist",
    "Hamming",
    "Jaro",
    "JaroWinkler",
    "Levenshtein",
    "LongestCommonSubstring",
    "OptimalStringAlignment",
    "SubstitutionMatrix",
    "TcrDist",
    "single_dispatch",
]



r"""
single_dispatch
===============

Functional implementations of the normal Metric classes, with the difference that the comparison happens only between
two strings. This can be useful for integration with other tools such as PySpark, where we want to access the fast
distance function implementations, without the forced pairwise comparisons.

Examples
--------

An example for computing the pairwise sequence distances using PySpark and setriq:

>>> from pyspark.sql import SparkSession
>>> from pyspark.sql.functions import udf
>>> from pyspark.sql.types import DoubleType
>>>
>>> spark = SparkSession \
...    .builder \
...    .master("local[2]") \
...    .appName("setriq-spark") \
...    .getOrCreate()
>>>
>>> df = spark.createDataFrame([('CASSLKPNTEAFF',), ('CASSAHIANYGYTF',), ('CASRGATETQYF',)], ['sequence'])
>>> df = df.withColumnRenamed('sequence', 'a').crossJoin(df.withColumnRenamed('sequence', 'b'))
>>>
>>> lev_udf = udf(levenshtein, returnType=DoubleType())  # single dispatch levenshtein distance
>>> df = df.withColumn('distance', lev_udf('a', 'b'))
>>> df.show()
+--------------+--------------+--------+
|             a|             b|distance|
+--------------+--------------+--------+
| CASSLKPNTEAFF| CASSLKPNTEAFF|     0.0|
| CASSLKPNTEAFF|CASSAHIANYGYTF|     8.0|
| CASSLKPNTEAFF|  CASRGATETQYF|     8.0|
|CASSAHIANYGYTF| CASSLKPNTEAFF|     8.0|
|CASSAHIANYGYTF|CASSAHIANYGYTF|     0.0|
|CASSAHIANYGYTF|  CASRGATETQYF|     9.0|
|  CASRGATETQYF| CASSLKPNTEAFF|     8.0|
|  CASRGATETQYF|CASSAHIANYGYTF|     9.0|
|  CASRGATETQYF|  CASRGATETQYF|     0.0|
+--------------+--------------+--------+

"""

from typing import List, Optional

import setriq._C as C

from .substitution import BLOSUM45, SubstitutionMatrix
from .utils import (
    check_jaro_weights,
    check_jaro_winkler_params,
    ensure_equal_sequence_length_sd,
    single_dispatch,
    tcr_dist_sd_component_check,
)

__all__ = [
    "cdr_dist",
    "levenshtein",
    "tcr_dist",
    "hamming",
    "jaro",
    "jaro_winkler",
    "longest_common_substring",
    "optimal_string_alignment",
]


@single_dispatch
def cdr_dist(
    a: str,
    b: str,
    substitution_matrix: SubstitutionMatrix = BLOSUM45,
    gap_opening_penalty: float = 10.0,
    gap_extension_penalty: float = 1.0,
) -> float:
    """
    Compute the CDRdist [1]_ metric between two sequences.

    {params}
    substitution_matrix: SubstitutionMatrix
        A substitution matrix object to inform the alignment scoring.
    gap_opening_penalty: float
        The penalty given to an alignment based on a gap opening. Values other than 0 give affine gap scoring.
        (default=10.0)
    gap_extension_penalty: float
        The penalty used to score the extension of the gap after opening. (default=1.0)

    {returns}

    Examples
    --------
    >>> seq = ('AASQ', 'PASQ')
    >>> cdr_dist(*seq)  # default params
    >>> cdr_dist(*seq, substitution_matrix=BLOSUM45,  # custom params
    ...          gap_opening_penalty=5.0, gap_extension_penalty=2.0)

    References
    ----------
    .. [1] Thakkar, N. and Bailey-Kellogg, C., 2019. Balancing sensitivity and specificity in distinguishing TCR groups
       by CDR sequence similarity. BMC bioinformatics, 20(1), pp.1-14. (https://doi.org/10.1186/s12859-019-2864-8)

    """
    distance = C.cdr_dist_sd(
        a,
        b,
        **substitution_matrix,  # type: ignore[arg-type]
        gap_opening_penalty=gap_opening_penalty,
        gap_extension_penalty=gap_extension_penalty
    )
    return distance


@single_dispatch
def levenshtein(a: str, b: str, extra_cost: float = 0.0) -> float:
    """
    Compute the Levenshtein distance [1]_ between two sequences. Based on the implementation in [2]_.

    {params}
    extra_cost: float
        Additional cost assigned by Levenshtein algorithm.

    {returns}

    Examples
    --------
    >>> levenshtein('AASQ', 'PASQ')

    References
    ----------
    .. [1] Levenshtein, V.I., 1966, February. Binary codes capable of correcting deletions, insertions, and reversals.
       In Soviet physics doklady (Vol. 10, No. 8, pp. 707-710).
    .. [2] python-Levenshtein (https://github.com/ztane/python-Levenshtein)

    """
    distance = C.levenshtein_sd(a, b, extra_cost=extra_cost)
    return distance


@single_dispatch
@ensure_equal_sequence_length_sd
def tcr_dist_component(
    a: str,
    b: str,
    substitution_matrix: SubstitutionMatrix,
    gap_penalty: float,
    gap_symbol: str = "-",
    weight: float = 1.0,
) -> float:
    distance = C.tcr_dist_component_sd(
        a,
        b,
        **substitution_matrix,  # type: ignore[arg-type]
        gap_penalty=gap_penalty,
        gap_symbol=gap_symbol,
        weight=weight
    )
    return distance


@tcr_dist_sd_component_check
def tcr_dist(a: dict, b: dict, **component_def) -> float:
    """
    Compute the TCRdist [1]_ metric between two sequences.

    Parameters
    ----------
    a: dict

    b: dict
    component_def

    Returns
    -------
    distance: float
        The distance between the two sequences.

    References
    ----------
    .. [1] Dash, P., Fiore-Gartland, A.J., Hertz, T., Wang, G.C., Sharma, S., Souquette, A., Crawford, J.C., Clemens,
       E.B., Nguyen, T.H., Kedzierska, K. and La Gruta, N.L., 2017. Quantifiable predictive features define
       epitope-specific T cell receptor repertoires. Nature, 547(7661), pp.89-93. (https://doi.org/10.1038/nature22383)

    """
    distance = 0.0
    for name, component in component_def.items():
        distance += tcr_dist_component(a[name], b[name], **component_def[name])
    return distance


@single_dispatch
@ensure_equal_sequence_length_sd
def hamming(a: str, b: str, mismatch_score: float = 1.0) -> float:
    """
    Compute the Hamming [1]_ distance between two sequences.

    {params}
    mismatch_score: float
        The weight given to a mismatch between the two sequence positions.

    {returns}

    Examples
    --------
    >>> hamming('AASQ', 'PASQ')
    >>> hamming('AASQ', 'PAS')  # error! different length sequences

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Hamming_distance

    """
    distance = C.hamming_sd(a, b, mismatch_score=mismatch_score)
    return distance


@single_dispatch
def jaro(a: str, b: str, jaro_weights: Optional[List[float]] = None) -> float:
    """
    Compute the Jaro [1] distance between two sequences. Adapted from [2].

    {params}
    jaro_weights: List[float]

    {returns}

    Examples
    --------
    >>> jaro('AASQ', 'PASQ')

    References
    ----------
    .. [1] Jaro, M.A., 1989. Advances in record-linkage methodology as applied to matching the 1985 census of Tampa,
       Florida. Journal of the American Statistical Association, 84(406), pp.414-420.
    .. [2] Van der Loo, M.P., 2014. The stringdist package for approximate string matching. R J., 6(1), p.111.

    """
    jaro_weights = check_jaro_weights(jaro_weights)
    distance = C.jaro_sd(a, b, jaro_weights=jaro_weights)
    return distance


@single_dispatch
@check_jaro_winkler_params
def jaro_winkler(
    a: str, b: str, p: float, max_l: int = 4, jaro_weights: Optional[List[float]] = None
) -> float:
    """
    Compute the Jaro-Winkler [1]_ distance between two sequences.

    {params}
    p: float
        The scaling factor applied to the common prefix re-weighting. The value needs to be in the range [0.0, 0.25]. If
        set to 0.0, Jaro-Winkler reduces down to Jaro.
    max_l: int
        The maximum length common prefix. (``default=4``)
    jaro_weights: List[float]

    {returns}

    Examples
    --------
    >>> jaro_winkler('AASQ', 'PASQ', p=0.10)

    References
    ----------
    .. [1] Winkler, W.E., 1990. String comparator metrics and enhanced decision rules in the Fellegi-Sunter model of
       record linkage.

    """
    jaro_weights = check_jaro_weights(jaro_weights)
    distance = C.jaro_winkler_sd(a, b, p=p, max_l=max_l, jaro_weights=jaro_weights)
    return distance


@single_dispatch
def longest_common_substring(a: str, b: str) -> float:
    """
    Compute the LCS distance [1]_ between two sequences.

    {params}

    {returns}

    Examples
    --------
    >>> longest_common_substring('AASQ', 'PASQ')

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Longest_common_substring_problem

    """
    distance = C.longest_common_substring_sd(a, b)
    return distance


@single_dispatch
def optimal_string_alignment(a: str, b: str) -> float:
    """
    Compute the OSA [1]_ between two sequences.

    {params}

    {returns}

    Examples
    --------
    >>> optimal_string_alignment('AASQ', 'PASQ')

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Damerau%E2%80%93Levenshtein_distance

    """
    distance = C.optimal_string_alignment_sd(a, b)
    return distance



from __future__ import annotations, print_function

import sys

if sys.version_info < (3,):
    print("Python 2 has reached end-of-life and is not supported by setriq.")
    sys.exit(-1)
if sys.platform == "win32" and sys.maxsize.bit_length() == 31:
    print(
        "32-bit Windows Python runtime is not supported. Please switch to 64-bit Python."
    )
    sys.exit(-1)

import logging
import pathlib
import platform
import re
import shutil
import subprocess
import traceback
from glob import glob
from typing import List

from pybind11.setup_helpers import ParallelCompile, Pybind11Extension
from setuptools import find_packages, setup

python_min_version = (3, 7, 2)
python_min_version_str = ".".join(map(str, python_min_version))
if sys.version_info < python_min_version:
    logging.error(
        "Minimum Python version required is {}. Current version is {}".format(
            python_min_version_str, platform.python_version()
        )
    )
    sys.exit(-1)

logging.basicConfig(
    level=logging.DEBUG,
    format="%(name)s - %(asctime)s.%(msecs)03d - %(levelname)s - %(module)s.%(funcName)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

# find the sem-rel version
DIR = pathlib.Path(__file__).parent

SOURCE_DIR = "src"
PROJECT_NAME = "setriq"
EXTENSION_NAME = "_C"

ParallelCompile("NPY_NUM_BUILD_JOBS").install()


class BuildFlags:
    _tools_formulae = {
        "Darwin": ("brew", "libomp"),
    }
    _args = {
        "Darwin": {"compiler": ["-Xpreprocessor", "-fopenmp"], "linker": ["-lomp"]}
    }
    _default_args = {"compiler": ["-fopenmp"], "linker": ["-fopenmp"]}
    compiler: list
    linker: list

    def __init__(self):
        self._system = platform.system()
        tool, formula = self._tools_formulae.get(self._system, ("apt", "libomp-dev"))
        args = self._args.get(self._system, self._default_args)

        not_found = self._libomp_check(tool, formula)
        if not_found is not None:
            logging.warning(
                f"{repr(not_found)} not found -- cannot compile parallelized code"
            )
            logging.info(
                "for information on how to enable CPU parallelization, "
                "please see https://github.com/BenTenmann/setriq#requirements"
            )
            for key in args:
                args[key] = []

        for key, val in args.items():
            self.__setattr__(key, val)

    @staticmethod
    def _libomp_check(tool, formula) -> None:
        if shutil.which(tool) is None:
            return tool

        formulae = subprocess.check_output([tool, "list"]).decode()
        if formula not in formulae:
            return formula

        return None


def get_requirements() -> List[str]:
    return (
        (pathlib.Path(".") / "requirements.txt").read_text().splitlines(keepends=False)
    )


def main():
    long_description = (DIR / "README.md").read_text()

    try:
        changelog = (DIR / "CHANGELOG.md").read_text()
        __version__, *_ = re.findall(r"\[([0-9.]+)]", changelog)
    except (FileNotFoundError, ValueError) as ex:
        __version__ = "0.1.0"
        logging.error(ex)
        logging.error(traceback.print_exc())
        logging.warning(
            f"Unable to get semantic release version. Setting version to {__version__}."
        )

    flags = BuildFlags()
    extensions = [
        Pybind11Extension(
            f"{PROJECT_NAME}.{EXTENSION_NAME}",
            sources=sorted(
                glob(
                    f"{SOURCE_DIR}/{PROJECT_NAME}/{EXTENSION_NAME}/**/*.cpp",
                    recursive=True,
                )
            ),
            cxx_std=14,
            define_macros=[
                ("VERSION_INFO", __version__),
                ("EXTENSION_NAME", EXTENSION_NAME),
            ],
            include_dirs=["include/setriq"],
            extra_compile_args=flags.compiler,
            extra_link_args=flags.linker,
        ),
    ]

    setup(
        name=PROJECT_NAME,
        version=__version__,
        description="Python package written in C++ for pairwise distance computation for sequences.",
        long_description=long_description,
        long_description_content_type="text/markdown",
        author="benjamin-tenmann",
        author_email="b.tenmann@me.com",
        url="https://github.com/BenTenmann/setriq",
        ext_modules=extensions,
        license="MIT",
        python_requires=">=3.7,<3.10",
        install_requires=get_requirements(),
        package_dir={f"{PROJECT_NAME}": f"{SOURCE_DIR}/{PROJECT_NAME}"},
        packages=find_packages(where=f"{SOURCE_DIR}", exclude=["tests", "scripts"]),
        package_data={f"{PROJECT_NAME}": ["data/*.json"]},
        include_package_data=True,
        classifiers=[
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Developers",
            "Topic :: Software Development :: Build Tools",
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
        ],
    )


if __name__ == "__main__":
    main()



"""
Package utilities. Not meant for outside use.
"""

import enum
import inspect
from functools import WRAPPER_ASSIGNMENTS, wraps
from typing import Callable, Dict, Iterable, List, Optional, Tuple, Union

from .substitution import BLOSUM62, SubstitutionMatrix

__all__ = [
    "enforce_list",
    "ensure_equal_sequence_length",
    "single_dispatch",
    "tcr_dist_sd_component_check",
    "ensure_equal_sequence_length_sd",
    "check_jaro_weights",
    "check_jaro_winkler_params",
    "TCR_DIST_DEFAULT",
    "TcrDistDef",
]

TcrDistComponentDef = Dict[str, Union[float, SubstitutionMatrix]]
NamedTCRDD = Tuple[str, TcrDistComponentDef]
TcrDistDef = List[NamedTCRDD]

WRAPPER_ASSIGNMENTS = (*WRAPPER_ASSIGNMENTS, "__signature__")  # type: ignore[assignment]
TCR_DIST_DEFAULT: TcrDistDef = [
    ("cdr_1", {"substitution_matrix": BLOSUM62, "gap_penalty": 4.0, "weight": 1.0}),
    ("cdr_2", {"substitution_matrix": BLOSUM62, "gap_penalty": 4.0, "weight": 1.0}),
    ("cdr_2_5", {"substitution_matrix": BLOSUM62, "gap_penalty": 4.0, "weight": 1.0}),
    ("cdr_3", {"substitution_matrix": BLOSUM62, "gap_penalty": 8.0, "weight": 3.0}),
]


class Argument(enum.Enum):
    POSITIONAL = 0
    KEYWORD = 1
    DEFAULT = 2


def _get_argument_index_from_argnum(n_params: int, argnum: int):
    # updates the argument index provided -- fixes issues with `_get_argument` when `argnum` < 0 (negative indexing)
    if argnum < 0:
        return argnum + n_params
    return argnum


def _get_argument(params, argname: str, argnum: int, args: tuple, kwargs: dict):
    # convenience function for getting a specific argument from all the args and kwargs passed to an arbitrary function
    # in Python, any argument can be passed as either positional or keyword (except if explicitly encoded otherwise)
    # and so we need to account for this, as well as the default case, when the argument is not passed
    if argname in kwargs:
        # we also return where we found the argument, i.e. the enum describes the argument type. This helps us
        # downstream, if we want to put an augmented argument back into args, kwargs
        return kwargs[argname], Argument.KEYWORD
    if len(args) > argnum:
        return args[argnum], Argument.POSITIONAL
    return params[argname].default, Argument.DEFAULT


def _put_argument(
    params,
    argname: str,
    argnum: int,
    arg_type: enum.Enum,
    argval,
    args: tuple,
    kwargs: dict,
):
    # this function is the complement to `_get_argument`
    if arg_type == Argument.POSITIONAL:
        args = tuple(arg if argnum != idx else argval for idx, arg in enumerate(args))
    elif arg_type == Argument.KEYWORD:
        kwargs[argname] = argval
    elif arg_type == Argument.DEFAULT:
        if params[argname].default != argval:
            kwargs[argname] = argval
    return args, kwargs


def _get_func_argument_info(fn: Callable, argnum: int):
    signature = inspect.signature(fn)
    params = signature.parameters
    argname = list(params)[argnum]

    return signature, params, argname


def _get_func_argument_info_from_name(fn: Callable, argname: str):
    signature = inspect.signature(fn)
    params = signature.parameters
    argnum = list(params).index(argname)

    return signature, params, argnum


def _add_func_signature(fn: Callable, signature: inspect.Signature) -> Callable:
    setattr(fn, "__signature__", getattr(fn, "__signature__", None) or signature)
    return fn


def enforce_list(argnum: int = 0, convert_iterable: bool = True):
    """
    Enforce that a specified argument is always passed as a list to a given function. This is a decorator factory.

    Parameters
    ----------
    argnum: int
        The positional (integer) index of the argument to be forced into list format. Works like regular positional
        indexing. (default=0)
    convert_iterable: bool
        Boolean defining whether to force convert any iterable (except ``str``) into a list.

    Returns
    -------
    decorator: Callable
        A new decorator function, which can be used to decorate an arbitrary function / method.

    Examples
    --------

    A basic example:

    >>> @enforce_list()
    ... def f(x):
    ...     return x * 3
    >>>
    >>> f([3])
    ... [3, 3, 3]
    >>> f(3)
    ... [3, 3, 3]

    Notice, that we ommit the ``argnum`` parameter here. This is because ``argnum`` is 0 by default, i.e. it looks at
    the first argument passed to ``f``. Note, that ``enforce_list`` needs to be called before decorating a function.

    Enforce list can also be composed arbitrarily, to enforce multiple arguments to be lists:

    >>> @enforce_list(argnum=0)
    ... @enforce_list(argnum=1)
    ... def f(x, y):
    ...     return x + y

    Finally, ``enforce_list`` can also force convert other iterables (excluding str) into lists:

    >>> @enforce_list(argnum=0, convert_iterable=True)
    ... def f(x):
    ...     return x * 3
    >>>
    >>> x = np.array([1.])
    >>> f(x)
    ... [1., 1., 1.]

    """
    if callable(argnum):
        raise TypeError(
            f"Make sure to call {repr(enforce_list.__name__)} before decorating."
        )

    def decorator(fn):
        signature, params, argname = _get_func_argument_info(fn, argnum)
        argidx = _get_argument_index_from_argnum(len(params), argnum)
        fn = _add_func_signature(fn, signature)

        @wraps(fn, assigned=WRAPPER_ASSIGNMENTS)
        def _fn(*args, **kwargs):
            argument, arg_type = _get_argument(params, argname, argidx, args, kwargs)
            if isinstance(argument, Iterable) and not isinstance(argument, str):
                if not isinstance(argument, list) and convert_iterable:
                    argument = list(argument)
            else:
                argument = [argname]
            args, kwargs = _put_argument(
                params, argname, argidx, arg_type, argument, args, kwargs
            )
            out = fn(*args, **kwargs)
            return out

        return _fn

    return decorator


def ensure_equal_sequence_length(argnum: int):
    """
    Ensure that all input sequences are of equal length.

    Parameters
    ----------
    argnum: int
        The positional (integer) index of the argument to be forced into list format. Works like regular positional
        indexing. (``default=0``)

    Returns
    -------
    decorator: Callable
        The decorator used to wrap the function where all sequences ought to be of equal length.

    Examples
    --------
    >>> @ensure_equal_sequence_length(argnum=0)
    ... def f(sequences):
    ...     return sequences
    >>>
    >>> a = ['AASQ', 'PWSQ']  # sequences of equal length
    >>> b = ['GAT', 'AAFFD']  # sequences with varying length
    >>>
    >>> f(a)  # no error
    >>> f(b)  # error!

    """
    import pandas as pd

    def decorator(fn):
        signature, params, argname = _get_func_argument_info(fn, argnum)
        argidx = _get_argument_index_from_argnum(len(params), argnum)
        fn = _add_func_signature(fn, signature)

        @wraps(fn, assigned=WRAPPER_ASSIGNMENTS)
        def _fn(*args, **kwargs):
            argument, _ = _get_argument(params, argname, argidx, args, kwargs)
            if (
                argument
                and not (len(argument[0]) == pd.Series(argument).str.len()).all()
            ):
                raise ValueError("Sequences must be of equal length")
            out = fn(*args, **kwargs)
            return out

        return _fn

    return decorator


def ensure_equal_sequence_length_sd(fn: Callable) -> Callable:
    signature = inspect.signature(fn)
    fn = _add_func_signature(fn, signature)

    add_doc = """Note
    ----
    `a` and `b` must be of equal length.
    """

    fn.__doc__ = (fn.__doc__ or "") + add_doc

    @wraps(fn, assigned=WRAPPER_ASSIGNMENTS)
    def _fn(a, b, *args, **kwargs):
        if len(a) != len(b):
            raise ValueError("Sequences must be of equal length")
        out = fn(a, b, *args, **kwargs)
        return out

    return _fn


def single_dispatch(fn: Callable) -> Callable:
    signature = inspect.signature(fn)
    fn = _add_func_signature(fn, signature)

    empty_doc = f"""
    Compute the `{fn.__name__}` metric between two sequences.

    {{params}}

    {{returns}}
    """

    param_doc = """
    Parameters
    ----------
    a: str
        A sequence to be compared.
    b: str
        A sequence to be compared."""

    return_doc = """Returns
    -------
    distance: float
        The computed distance between the two sequences."""

    fn.__doc__ = (fn.__doc__ or empty_doc).format(params=param_doc, returns=return_doc)

    @wraps(fn, assigned=WRAPPER_ASSIGNMENTS)
    def _fn(a, b, *args, **kwargs):
        if any(not isinstance(sequence, str) for sequence in (a, b)):
            raise TypeError("`a` and `b` must be of type str")
        out = fn(a, b, *args, **kwargs)
        return out

    return _fn


def tcr_dist_sd_component_check(fn):
    # wrapper specifically used for the single dispatch tcr_dist function to check inputs (components)
    import os

    signature = inspect.signature(fn)
    fn = _add_func_signature(fn, signature)

    def _check_component(name, component: dict):
        essential_keys = ["substitution_matrix", "gap_penalty"]
        optional_keys = ["gap_symbol", "weight"]
        missing_keys = set(essential_keys).difference(component)
        if missing_keys:
            msg = ", ".join(map(repr, missing_keys))
            raise ValueError(f"missing keys in component def {repr(name)}: {msg}")

        init_types: List[Union[type, Tuple[type, ...]]] = [
            SubstitutionMatrix,
            (float, int),
            str,
            (float, int),
        ]
        for key, _type in zip(essential_keys + optional_keys, init_types):
            elem = component.get(key)
            if elem is not None and not isinstance(elem, _type):
                given_type = type(elem)
                raise TypeError(
                    f"{repr(key)} needs to be of type {repr(_type)}, not {repr(given_type)}"
                )

    @wraps(fn, assigned=WRAPPER_ASSIGNMENTS)
    def _fn(a, b, **component_def):
        if not os.environ.get("SKIP_TCR_DIST_COMPONENT_CHECK"):
            for name, component in component_def.items():
                _check_component(name, component)
            if not component_def:
                component_def = dict(TCR_DIST_DEFAULT)
            if set(component_def).difference(set(a).union(b)):
                raise ValueError(
                    "key mismatch between payloads (`a` and `b`) and defined components."
                )
        out = fn(a, b, **component_def)
        return out

    return _fn


def check_jaro_weights(weights: Optional[List[float]]) -> List[float]:
    if weights is None:
        weights = [1 / 3] * 3
    if len(weights) != 3:
        raise ValueError("`jaro_weights` has to be of length 3")
    if sum(weights) != 1.0:
        raise ValueError("`jaro_weights` has to sum to 1.0")
    return weights


def check_jaro_winkler_params(fn: Callable):
    # checks that Jaro-Winkler parameters are sensibly defined
    argname_p = "p"
    argname_l = "max_l"
    signature, params, argnum_p = _get_func_argument_info_from_name(fn, argname_p)
    _, _, argnum_l = _get_func_argument_info_from_name(fn, argname_l)
    fn = _add_func_signature(fn, signature)

    @wraps(fn, assigned=WRAPPER_ASSIGNMENTS)
    def _fn(*args, **kwargs):
        p, _ = _get_argument(params, argname_p, argnum_p, args, kwargs)
        max_l, _ = _get_argument(params, argname_l, argnum_l, args, kwargs)

        if not (0.0 <= p <= 0.25):
            raise ValueError("`p` must be in range [0.0, 0.25]")
        if max_l < 0 or not isinstance(max_l, int):
            raise ValueError("`max_l` must be a non-negative integer")
        out = fn(*args, **kwargs)
        return out

    return _fn
